// IRCode.cpp: implementation of the CIRCode class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "IRTool.h"
#include "IRCode.h"
#include "DecodeIR.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CIRCode::CIRCode()
{
	m_data.SetSize(50, 1);	// NEC is largest with 34 pairs for onetime and repeat
							// although sometimes I hack the codes for an extra
							// 16 pairs

}

CIRCode::~CIRCode()
{
	m_data.RemoveAll();
}

IRCODE_FORMAT 
CIRCode::WhatFormat()
{
	return m_CurrentFormat;
}




int 
CIRCode::DecodeHex(LPCTSTR hex, CCode& code)
{
	WORD	i, total_length;
	WORD	word;
	WORD	word2;
	char *token;
	char *p = new TCHAR[strlen(hex) + 1];
	

	m_CurrentFormat = IR_UNKNOWN;

	_tcscpy(p, hex);
	token = strtok(p, " \t\n");
	if ((0 == token) || (sscanf(token, "%lx", &word) < 1))
	{
		m_nError = IDS_INVALID_CODE;
		delete [] p;
		return -m_nError;
	}

	if (word)
	{
		m_nError = IDS_NOT_LEARNED;
		delete [] p;
		return -m_nError;
	}

	token = strtok(NULL, " \t\n");
	if ((0 == token) || (sscanf(token, "%lx", &word) < 1))
	{
		m_nError = IDS_NO_CARRIER;
		delete [] p;
		return -m_nError;

	}
	code.m_nCarrier = m_nCarrier = word;

	token = strtok(NULL, " \t\n");
	if ((0 == token) || (sscanf(token, "%lx", &word) < 1))
	{
		m_nError = IDS_NO_OT_LENGTH;
		delete [] p;
		return -m_nError;

	}
	code.m_bOneTime = (word != 0);
	m_nOneTimeLength = word;

	token = strtok(NULL, " \t\n");
	if ((0 == token) || (sscanf(token, "%lx", &word) < 1))
	{
		m_nError = IDS_NO_REP_LENGTH;
		delete [] p;
		return -m_nError;

	}
	code.m_bRepeat = (word != 0);
	m_nRepeatLength = word;

	if (0 == m_nOneTimeLength && 0 == m_nRepeatLength)
	{
		m_nError = IDS_ZERO_LENGTH;
		delete [] p;
		return -m_nError;

	}

	m_data.SetSize(m_nOneTimeLength + m_nRepeatLength);


	total_length = m_nOneTimeLength + m_nRepeatLength;
	for (i = 0; i < total_length; i++)
	{
		token = strtok(NULL, " \t\n");
		if ((0 == token) || (sscanf(token, "%lx", &word) < 1))
		{
			m_nError = IDS_INCOMPLETE_CODE;
			delete [] p;
			return -m_nError;
			
		}
		token = strtok(NULL, " \t\n");
		if ((0 == token) || (sscanf(token, "%lx", &word2) < 1))
		{
			m_nError = IDS_INCOMPLETE_CODE;
			delete [] p;
			return -m_nError;
			
		}
		m_data[i].pair[0] = word;
		m_data[i].pair[1] = word2;
	}
	delete [] p;

	if (m_nError = CalculateBits() < 0)
	{
		return -m_nError;
	}

	code.m_bpZero.value = m_bpZero.value;
	code.m_bpOne.value = m_bpOne.value;

	//Call decoder DLL 
	HINSTANCE hDll = NULL;
	hDll = LoadDll();
	if (hDll != NULL)
		TryDLL(m_data,m_nCarrier,m_nOneTimeLength,m_nRepeatLength);

	if (hDll != NULL)
		FreeLibrary(hDll);

	//End Call Decoder DLL

	if (m_nOneTimeLength > 0)
	{
		DecipherPairs(code.m_cpOneTime, m_nOneTimeLength, 0, FALSE);

	}

	if (m_nRepeatLength > 0)
	{
		DecipherPairs(code.m_cpRepeat, m_nRepeatLength, m_nOneTimeLength, TRUE);

	}
	code.m_type = m_CurrentFormat;
	return m_CurrentFormat;
}


//
// Given an IR Code, figure out the likely ones and zeroes
//
int 
CIRCode::CalculateBits()
{
	Histogram	*hist;
	int error;

	error = 0;

	/*
	 * First smooth the pairs out so we can generate the histogram based
	 * on normalized input.
	 */
	SmoothValues();
	/*
	 * Now count up the various normalized values.
	 */
	hist = CountValues();
	if (hist == NULL || hist->next == NULL)
	{
		error = -IDS_TOO_FEW_VALUES;
	}

	/*
	 * The value for "1" is the greater of the two most frequent values,
	 * and "0" is the lesser of the two.  If there's ambiguity as to which
	 * values are the most frequent, it's an error.
	 */
	else if (hist->next->next != NULL &&
	    hist->next->next->value == hist->next->value)
	{
		error =  -IDS_AMBIGUOUS;
	}

	else if (hist->value > hist->next->value)
	{
		m_bpZero.value = hist->next->value;
		m_bpOne.value = hist->value;
	}
	else
	{
		m_bpZero.value = hist->value;
		m_bpOne.value = hist->next->value;
	}

	FreeHistogram(hist);

	return error;
}

//
// Smooth out burst values to make decoding easier.  Since learned codes are
// based on sampling of a physical signal at a finite sample rate, the values
// can be slightly variable, e.g. a burst pair of 001A 003E is probably the
// same thing as 001B 003D.  We treat two burst values as the same if they're
// within a configurable distance (the "fuzz factor") of each other.
//
// We use a naive O(n log n) algorithm: for each value X, scan forward in the
// list of pairs and, if a subsequent value Y is approximately equal to X,
// set it precisely equal to X.  There is probably a slightly faster way to
// do this.

void 
CIRCode::SmoothValues()
{
	
	CBurstPair * data = m_data.GetData();

	WORD length = (m_nOneTimeLength + m_nRepeatLength) * 2;


	for (WORD i = 0; i < length; i++)
	{
		for (WORD j = i + 1; j < length; j++)
		{
			if (abs(data->pair[i] - data->pair[j]) < FUZZ_FACTOR)
			{
				data->pair[j] = data->pair[i];

				/*
				 * Avoid a little bit of redundant scanning
				 * by advancing the starting position each
				 * time it's on the first of two identical
				 * values.  This has the effect of skipping
				 * over blocks of identical values.
				 */
				if (j == i + 1)
					i++;
			}
		}
	}
}


//
// Generate a histogram of burst-pair values in the data set.  This will be
// used to determine which two values are most likely to represent "1" and
// "0".
//
Histogram * 
CIRCode::CountValues()
{
	Histogram	*head = NULL, *tail = NULL, *cur, *temp;

	for (WORD i = 0; i < m_nOneTimeLength + m_nRepeatLength; i++)
	{
		/*
		 * Is there an existing histogram entry for this value?
		 */
		for (cur = head; cur != NULL; cur = cur->next)
			if (cur->value == m_data[i].value)
				break;
		if (cur != NULL)
		{
			/*
			 * Yes.  Increment its count and move it forward in
			 * the list if it's more frequent than its predecessor.
			 */
			cur->count++;

			if (cur->prev != NULL &&
			    cur->prev->count < cur->count)
			{
				/* Unlink from the old position */
				temp = cur->prev;
				temp->next = cur->next;
				if (cur->next != NULL)
					cur->next->prev = temp;
				else
					tail = temp;

				/* And insert into the new */
				cur->prev = temp->prev;
				cur->next = temp;
				if (temp->prev != NULL)
					temp->prev->next = cur;
				else
					head = cur;
				temp->prev = cur;
			}

			continue;
		}

		/*
		 * This is a new value.  Make a histogram entry for it and put
		 * it at the end of the list.
		 */
		cur = new Histogram ;
		if (cur == NULL)
		{
			FreeHistogram(head);
			m_nError = IDS_HIST_OUT_OF_MEMORY;
			head = NULL;
			return head;
		}
		else
		{
			cur->value = m_data[i].value;
			cur->count = 1;
			cur->next = NULL;
			cur->prev = tail;
			if (tail != NULL)
				tail->next = cur;
			else
				head = cur;
			tail = cur;
		}
	}

	return (head);
}

//
// Free the histogram
//
void 
CIRCode::FreeHistogram(Histogram *head)
{
	Histogram	*next;

	while (head != NULL)
	{
		next = head->next;
		delete head;
		head = next;
	}
}



//
// Pull an integer of some number of bits (up to 32) from a burst-pair
// sequence in reverse bit order.
//
DWORD 
CIRCode::GrabInteger(int start, int length)
{
	DWORD	thisbit = 1;
	DWORD	number = 0;

	while (length--)
	{
		if (m_data[start++].value == m_bpOne.value)
			number |= thisbit;
		thisbit <<= 1;
	}

	return number;
}

//
// Pull an integer of some number of bits (up to 32) from a burst-pair
// sequence in reverse bit order.  Includes the lead-out in the calculation of the
// last bit.
//
DWORD 
CIRCode::GrabIntegerSony(int start, int length)
{
	DWORD		thisbit = 1;
	DWORD		number = 0;
	CBurstPair	leadout;

	
	leadout.value = m_data[start + length - 1].value;
	
	while (length--)
	{
		if (m_data[start].value == m_bpOne.value)
			number |= thisbit;
		else if ((m_data[start].value == leadout.value) && (m_data[start].pair[0] == m_bpOne.pair[0]))
			number |= thisbit;
		thisbit <<= 1;
		start++;
	}

	return number;
}



//
// Interpret a bit definition, which is always two hex numbers.
//
void 
CIRCode::InterpretBit(LPCTSTR str, CBurstPair& val)
{
	WORD			on, off;
	LPCTSTR			line = str;

	while (isspace(line[0]))
		line++;
	if (sscanf(line, "%lx %lx", &on, &off) < 2)
	{
		m_nError = IDS_NO_BURST_PAIR;
	}
	else
	{
		val.pair[0] = on;
		val.pair[1] = off;

	}

}




// Interpret a code line.  This is either a string of pair values (where a
// pair value can be two hex numbers or "0" or "1") or a human-readable
// code specification.
//
// is_repeat should be true if this line specifies a code's repeating part.
//
int 
CIRCode::InterpretCode(CCode& code)
{
	int	length = 0;
	int	length2 = 0;

	m_data.RemoveAll();
	CString str2;
	code.m_cpOneTime.m_nLength = code.m_cpRepeat.m_nLength = 0;

	switch (code.m_type)
	{
	case IR_NEC:
		length = InterpretCodeNEC(code);
		break;
	case IR_DENON:
		length = InterpretCodeDENON(code);
		break;

	case IR_JVC:
		length = InterpretCodeJVC(code);
		break;
	case IR_KENWOOD:
		length = InterpretCodeKenwood(code);
		break;
	case IR_SONY_5:
	case IR_SONY_8:
	case IR_SONY_13:
		length = InterpretCodeSONY(code);
		break;
	default:
		m_data.SetSize(128);
		length = InterpretCodeHEX(code, 0, false);
		if (length < 0)
			return length;
		code.m_cpOneTime.m_nLength = length;
		length2 = InterpretCodeHEX(code, length, true);
		if (length2 < 0)
			return length2;
		code.m_cpRepeat.m_nLength = length2;
	}
	return length + length2;

}


//
// Interpret a pretty-printed DENON code.
//
int 
CIRCode::InterpretCodeDENON(CCode& code)
{
	WORD	index = 0;
	DWORD	command = 0;
	DWORD	device = 0;
	DWORD	temp;

	/*
	 * DENON  codes max out to 32 pairs
	 * Format:  device(5), command(10), pad, device(5), ~command(10), leadout
	 */
	m_data.SetSize(64);


	if (code.m_bOneTime)
	{
		device = code.m_cpOneTime.m_nDevice & 0x1F;
		command = code.m_cpOneTime.m_nCommand & 0x3FF;
		EncodeValue(device, index, 5);
		EncodeValue(command, index, 10);
		if (code.m_cpOneTime.m_bpPad.value != 0)
		{
			m_data[index++].value = code.m_cpOneTime.m_bpPad.value;
			EncodeValue(device, index, 5);
			EncodeValue(~command, index, 10);
		}
		m_data[index++].value = code.m_cpOneTime.m_bpLeadOut.value;
		code.m_cpOneTime.m_nLength = index ;
	}

	if (code.m_bRepeat)
	{

		device = code.m_cpRepeat.m_nDevice & 0x1F;
		command = code.m_cpRepeat.m_nCommand & 0x3FF;
		temp = command;
		if (code.m_cpOneTime.m_nLength != 0)
			temp = (~temp & 0x3ff);
		EncodeValue(device, index, 5);
		EncodeValue(temp, index, 10);
		if (code.m_cpRepeat.m_bpPad.value != 0)
		{
			temp = command;
			if (code.m_cpOneTime.m_nLength == 0)
				temp = (~temp & 0x3ff);
			m_data[index++].value = code.m_cpRepeat.m_bpPad.value;
			EncodeValue(device, index, 5);
			EncodeValue(temp, index, 10);
		}
		m_data[index++].value = code.m_cpRepeat.m_bpLeadOut.value;
		code.m_cpRepeat.m_nLength = index - code.m_cpOneTime.m_nLength;
	}
	return index;
}


//
// Interpret a hex-dumped code line.
//
int 
CIRCode::InterpretCodeHEX(CCode& code, WORD start, BOOL bIsRepeat)
{
	char	*token;
	WORD	num;
	int		index;

	char*	line = new TCHAR[code.m_cpOneTime.m_sRaw.GetLength() +
		code.m_cpRepeat.m_sRaw.GetLength() + 1];

	/*
	 * If every entry on the line is a bit specification, we'll need
	 * one burst pair times half the number of characters in the line
	 * (since the bits are separated by whitespace).  Allocate enough
	 * memory for that worst case.
	 */
	if (bIsRepeat)
		_tcscpy(line, LPCTSTR(code.m_cpRepeat.m_sRaw));
	else
		_tcscpy(line, LPCTSTR(code.m_cpOneTime.m_sRaw));

	index = start;

	/*
	 * Now walk through all the entries on the line.
	 */

	token = strtok(line, "\n\t ");
	while (token != NULL)
	{
		/*
		 * If an entry is a bit, put the appropriate pair into the
		 * stream.
		 */
		if (! strcmp(token, "0"))
			m_data[index].value = m_bpZero.value;
		else if (! strcmp(token, "1"))
			m_data[index].value = m_bpOne.value;
		else
		{
			/*
			 * This plus the next entry is a burst pair.
			 */
			if (sscanf(token, "%lx", &num) < 1)
			{
				m_nError = IDS_CANT_DECODE;
				delete [] line;
				return -m_nError;
			}

			m_data[index].pair[0] = num;

			token = strtok(NULL, "\n\t ");
			if ((0 == token) ||
				sscanf(token, "%lx", &num) < 1)
			{
				m_nError = IDS_CANT_DECODE;
				delete [] line;
				return -m_nError;
			}

			m_data[index].pair[1] = num;
		}

		index++;
		token = strtok(NULL, "\n\t ");
	}
	delete [] line;
	return index - start;

}


//
// Interpret a pretty-printed JVC code.
//
int 
CIRCode::InterpretCodeJVC(CCode& code)
{

	WORD	index = 0;
	DWORD	command = 0;
	DWORD	device = 0;

	/*
	 * JVC codes are  to 18  pairs long.
	 */
	m_data.SetSize(36);


	if (code.m_bOneTime)
	{
		if (code.m_cpOneTime.m_bpLeadIn.value != 0)
			m_data[index++].value = code.m_cpOneTime.m_bpLeadIn.value;
		device = code.m_cpOneTime.m_nDevice & 0xFF;
		command = code.m_cpOneTime.m_nCommand & 0xFF;
		EncodeValue(device, index, 8);
		EncodeValue(command, index, 8);
		if (code.m_cpOneTime.m_bpLeadOut.value != 0)
			m_data[index++].value = code.m_cpOneTime.m_bpLeadOut.value;
		code.m_cpOneTime.m_nLength = index ;
	}
	if (code.m_bRepeat)
	{
		if (code.m_cpRepeat.m_bpLeadIn.value != 0)
			m_data[index++].value = code.m_cpRepeat.m_bpLeadIn.value;
		device = code.m_cpRepeat.m_nDevice & 0xFF;
		command = code.m_cpRepeat.m_nCommand & 0xFF;
		EncodeValue(device, index, 8);
		EncodeValue(command, index, 8);
		if (code.m_cpRepeat.m_bpLeadOut.value != 0)
			m_data[index++].value = code.m_cpRepeat.m_bpLeadOut.value;
		code.m_cpRepeat.m_nLength = index - code.m_cpOneTime.m_nLength;
	}

	return index;
}

//
// Interpret a pretty printed NEC code
//
int 
CIRCode::InterpretCodeNEC(CCode& code)
{

	WORD	index = 0;
	DWORD	command = 0;
	DWORD	device = 0;

	/*
	 * NEC codes are always 34 pairs long.
	 */
	m_data.SetSize(68);

	/*
	 * Walk through the entries on the line and interpret each one.
	 */

	
	if (code.m_bOneTime)
	{
		m_data[index++].value = code.m_cpOneTime.m_bpLeadIn.value;
		device = (((~code.m_cpOneTime.m_nDevice) & 0xFF) << 8) | (code.m_cpOneTime.m_nDevice & 0xFF);
		command = (((~code.m_cpOneTime.m_nCommand) & 0xFF) << 8) | (code.m_cpOneTime.m_nCommand & 0xFF);
		EncodeValue(device, index, 16);
		EncodeValue(command, index, 16);
		m_data[index++].value = code.m_cpOneTime.m_bpLeadOut.value;
		code.m_cpOneTime.m_nLength = index;
	}

	if (code.m_bRepeat)
	{
		m_data[index++].value = code.m_cpRepeat.m_bpLeadIn.value;
		device = (((~code.m_cpRepeat.m_nDevice) & 0xFF) << 8) | (code.m_cpRepeat.m_nDevice & 0xFF);
		command = (((~code.m_cpRepeat.m_nCommand) & 0xFF) << 8) | (code.m_cpRepeat.m_nCommand & 0xFF);
		EncodeValue(device, index, 16);
		EncodeValue(command, index, 16);
		m_data[index++].value = code.m_cpRepeat.m_bpLeadOut.value;
		code.m_cpRepeat.m_nLength = index - code.m_cpOneTime.m_nLength;
	}
	return index;
}


//
// Interpret pretty pringed SONY code
//
int 
CIRCode::InterpretCodeSONY(CCode& code)
{

	WORD index = 0;
	WORD devlen = 0;
	DWORD command = 0;
	DWORD device = 0;

	/*
	 * SONY codes are 12 to 20  pairs long.
	 */
	m_data.SetSize(40);

	switch (code.m_type)
	{
	case IR_SONY_5:
		devlen = 5;
		break;
	case IR_SONY_8:
		devlen = 8;
		break;
	case IR_SONY_13:
		devlen = 13;
		break;
	}

	if (code.m_bOneTime)
	{
		m_data[index++].value = code.m_cpOneTime.m_bpLeadIn.value;
		device = code.m_cpOneTime.m_nDevice & 0x1FFF;
		command = code.m_cpOneTime.m_nCommand & 0x7F;
		EncodeValue(command, index, 7);
		EncodeValue(device, index, devlen);
		m_data[index - 1].pair[1] = code.m_cpOneTime.m_bpLeadOut.pair[1];
		code.m_cpOneTime.m_nLength = index;
	}

	if (code.m_bRepeat)
	{
		m_data[index++].value = code.m_cpRepeat.m_bpLeadIn.value;
		device = code.m_cpRepeat.m_nDevice & 0x1FFF;
		command = code.m_cpRepeat.m_nCommand & 0x7F;
		EncodeValue(command, index, 7);
		EncodeValue(device, index, devlen);
		m_data[index - 1].pair[1] = code.m_cpRepeat.m_bpLeadOut.pair[1];
		code.m_cpRepeat.m_nLength = index - code.m_cpOneTime.m_nLength ;
	}
	return index;
}



int
CIRCode::DecodeHex(const WORD hex[], WORD length, CCode& code)
{

	if (length <= 4 && 	(length < (hex[2] + hex[3] + 2)))
	{
		m_nError = IDS_TOO_FEW_VALUES;
		return -m_nError;
	}


	if (hex[0])
	{

		m_nError = IDS_NOT_LEARNED;
		return -m_nError;
	}

	m_nCarrier = hex[1];
	m_nOneTimeLength = hex[2];
	m_nRepeatLength = hex[3];

	if (0 == m_nOneTimeLength && 0 == m_nRepeatLength)
	{
		m_nError = IDS_INVALID_CODE;
		return -m_nError;
	}

	m_data.SetSize(m_nOneTimeLength + m_nRepeatLength);


	WORD wTotalLength = m_nOneTimeLength + m_nRepeatLength;
	for (WORD i = 0; i < wTotalLength; i++)
	{
		m_data[i].pair[0] = hex[i * 2 + 4];
		m_data[i].pair[1] = hex[i * 2 + 5];

	}

		if (m_nError = CalculateBits() < 0)
	{
		return -m_nError;
	}

	code.m_bpZero.value = m_bpZero.value;
	code.m_bpOne.value = m_bpOne.value;

	if (m_nOneTimeLength > 0)
	{
		DecipherPairs(code.m_cpOneTime, m_nOneTimeLength, 0, FALSE);

	}

	if (m_nRepeatLength > 0)
	{
		DecipherPairs(code.m_cpRepeat, m_nRepeatLength, m_nOneTimeLength, TRUE);

	}
	code.m_type = m_CurrentFormat;
	return m_CurrentFormat;
}




int 
CIRCode::DecipherPairs(CCodePart &cp, WORD length, WORD start, BOOL bIsRepeat)
{

	if (IsNEC(length, start, bIsRepeat))
		DecipherPairsNEC(cp, length, start, bIsRepeat);
	else if (IsDENON(length, start, bIsRepeat))
		DecipherPairsDENON(cp, length, start, bIsRepeat);
	else if (IsSONY(length, start, bIsRepeat))
		DecipherPairsSONY(cp, length, start, bIsRepeat);
	else if (IsJVC(length, start, bIsRepeat))
		DecipherPairsJVC(cp, length, start, bIsRepeat);
	else if (IsKenwood(length, start, bIsRepeat))
		DecipherPairsKenwood(cp, length, start, bIsRepeat);
	else
		DecipherPairsRaw(cp, length, start, bIsRepeat);
	return m_CurrentFormat;
}

//
// Examine a burst-pair sequence to try to determine if it's in NEC format
// (leadin, 8 bit device code, complement of device code, 8 bit command,
// complement of command, leadout).
//
BOOL 
CIRCode::IsNEC(WORD length, WORD start, BOOL bIsRepeat)
{
	WORD	i;

	/* NEC format always has one leadin and one leadout pair. */
	if (length != 34)
		return FALSE;


	/* The interior is always all ones and zeros. */
	for (i = start + 1; i < start + 33; i++)
	{
		if (m_data[i].value != m_bpZero.value && m_data[i].value != m_bpOne.value)
			return FALSE;
	}
	/* See if the device code is properly complemented. */
	for (i = start + 1; i < start + 9; i++)
	{
		if (m_data[i].value == m_bpZero.value && m_data[i + 8].value == m_bpZero.value ||
		    m_data[i].value == m_bpOne.value && m_data[i + 8].value == m_bpOne.value)
			return FALSE;
	}

	/*
	 * See if the command code is properly complemented.
	 */
	for (i = start + 17; i < start + 25; i++)
	{
		if (m_data[i].value == m_bpZero.value && m_data[i + 8].value == m_bpZero.value ||
		    m_data[i].value == m_bpOne.value && m_data[i + 8].value == m_bpOne.value)
			return FALSE;
	}

	m_CurrentFormat = IR_NEC;

	return TRUE;
}

//
// Examine a burst-pair sequence to try to determine if it's in DENON format
// (5 bit device code, 10 bit command, (optional pad,
// device code, complement of command), leadout)
//
BOOL 
CIRCode::IsDENON(WORD length, WORD start, BOOL bIsRepeat)
{
	WORD	i;


	/* DENON format always has and one leadout pair or pad, command and leadout. */
	if ((length != 16) && (length != 32))
		return FALSE;

	/* The interior is always all ones and zeros. */
	for (i = start; i < start + 15 ; i++)
	{
		if (m_data[i].value != m_bpZero.value && m_data[i].value != m_bpOne.value)
			return FALSE;

	}
	m_CurrentFormat = IR_DENON;

	return TRUE;
}

//
// Examine a burst-pair sequence to try to determine if it's in JVC format
// (0012, 0011, leadin, 8 bit device code, 8 bit command,
// leadout).
//
BOOL 
CIRCode::IsJVC(WORD length, WORD start, BOOL bIsRepeat)
{
	WORD	i;


	/* JVC format always has one leadin and one leadout pair. or only leadout */
	if ((length != 0x12) && (length != 0x11))
		return FALSE;

	/* The interior is always all ones and zeros. */
	for (i = start + 1; i < start + length - 1; i++)
	{
		if (m_data[i].value != m_bpZero.value && m_data[i].value != m_bpOne.value)
			return FALSE;
	}
	m_CurrentFormat = IR_JVC;

	return TRUE;

}

//
// Examine a burst-pair sequence to try to determine if it's in SONY format
// (leadin, 7 bit command code, 5, 8 or 13 bit device code,
// leadout).
//
BOOL 
CIRCode::IsSONY(WORD length, WORD start, BOOL bIsRepeat)
{
	WORD	i;


	/* SONY format always has one leadin and one leadout pair. */
	/* But the leadout pair is counted as part of the device */
	if ((length != 13) && (length != 16) && (length != 21))
		return FALSE;

	//
	// Make sure that there is a leadin.  If the leadin is a zero or one
	// then this isn't sony.
	//
	if (m_data[start].value == m_bpZero.value || m_data[start].value == m_bpOne.value)
	{
		return FALSE;
	}

	/* The interior is always all ones and zeros. */
	/* However, ignore the last pair for now */
	for (i = start + 1; i < start + length - 1; i++)
	{
		if (m_data[i].value != m_bpZero.value && m_data[i].value != m_bpOne.value)
			return FALSE;
	}
	/* Ensure that the last pair is X + Y where X is a 0 or 1 high part of pair */
	if ((m_data[length - 1].pair[0] != m_bpZero.pair[0]) && 
		(m_data[length - 1].pair[0] != m_bpOne.pair[0]))
		return FALSE;

	if (13 == length)
		m_CurrentFormat = IR_SONY_5;
	else if (16 == length)
		m_CurrentFormat = IR_SONY_8;
	else
		m_CurrentFormat = IR_SONY_13;

	return TRUE;
}

void 
CIRCode::DecipherPairsNEC(CCodePart &cp, WORD length, WORD start, BOOL bIsRepeat)
{
	cp.m_bpLeadIn.value = m_data[start].value;
	cp.m_nDevice = GrabInteger(start + 1, 8);
	cp.m_nCommand = GrabInteger(start + 17, 8);
	cp.m_bpLeadOut.value = m_data[start + 33].value;
	cp.m_bpPad.value = 0;
	cp.m_sRaw = _T("");
}

void 
CIRCode::DecipherPairsDENON(CCodePart &cp, WORD length, WORD start, BOOL bIsRepeat)
{
	DWORD command = GrabInteger(start + 5, 10);
	//
	// If the onetime length nis not zero, then the command or
	// at least the first instance of it is the complement
	// Natch, we only care about this for the repeat part (if any)
	//

	if (bIsRepeat && (m_nOneTimeLength != 0))
	{
		command = (~command & 0x3ff);
	}
	cp.m_nDevice = GrabInteger(start, 5);
	cp.m_nCommand = command;
	cp.m_sRaw = _T("");
	cp.m_bpLeadIn.value = 0;
	if (length > 16)
	{
		cp.m_bpPad.value = m_data[start + 15].value;
		cp.m_bpLeadOut.value = m_data[start + 31].value;

	}
	else
	{
		cp.m_bpLeadOut.value = m_data[start + 15].value;
		cp.m_bpPad.value = 0;
	}
}

void 
CIRCode::DecipherPairsSONY(CCodePart& cp, WORD length, WORD start, BOOL bIsRepeat)
{
	cp.m_nDevice = GrabIntegerSony(start + 8, length - 8);
	cp.m_nCommand = GrabInteger(start + 1, 7);
	cp.m_bpLeadIn.value = m_data[start].value;
	cp.m_bpLeadOut.pair[0] = 0;
	cp.m_bpLeadOut.pair[1] = m_data[start + length - 1].pair[1];
	cp.m_bpPad.value = 0;
	cp.m_sRaw = _T("");
}

void 
CIRCode::DecipherPairsJVC(CCodePart &cp, WORD length, WORD start, BOOL bIsRepeat)
{
	cp.m_bpPad.value = 0;
	cp.m_sRaw = _T("");
	if ((m_data[start].value != m_bpOne.value) && (m_data[start].value != m_bpZero.value))
	{
		cp.m_bpLeadIn.value = m_data[start].value;
		cp.m_nDevice = GrabInteger(start + 1, 8);
		cp.m_nCommand = GrabInteger(start + 9, 8);
		cp.m_bpLeadOut.value = m_data[start + length - 1].value;
	}
	else
	{
		cp.m_bpLeadIn.value = 0;
		cp.m_nDevice = GrabInteger(start, 8);
		cp.m_nCommand = GrabInteger(start + 8, 8);
		cp.m_bpLeadOut.value = m_data[start + length - 1].value;
	}
}

void 
CIRCode::DecipherPairsRaw(CCodePart &cp, WORD length, WORD start, BOOL bIsRepeat)
{
	WORD	i;
	CString str2;
	cp.m_bpLeadIn.value = 0;
	cp.m_bpLeadOut.value = 0;
	cp.m_nCommand = 0;
	cp.m_bpPad.value = 0;
	cp.m_nDevice = 0;
	cp.m_sRaw = _T("");
	for (i = start; i < start + length; i++)
	{
		if (m_data[i].value == m_bpZero.value)
			cp.m_sRaw += "0 ";
		else if (m_data[i].value == m_bpOne.value)
			cp.m_sRaw += "1 ";
		else
		{
			str2.Format("%04x %04x ", m_data[i].pair[0], m_data[i].pair[1]);
			cp.m_sRaw += str2;
		}
	}
}

int 
CIRCode::EncodeHex(CString &hex, CCode &code)
{

	m_nError = 0;
	int length = InterpretCode(code);
	if (length > 0)
	{
		CString		str;
		hex.Format("0000 %04x %04x %04x", code.m_nCarrier, code.m_cpOneTime.m_nLength,
			code.m_cpRepeat.m_nLength);
		for (int i = 0; i < length; i++)
		{
			str.Format(" %04x %04x", 
				m_data[i].pair[0],
				m_data[i].pair[1]);
			hex += str;
		}
	}

	return -m_nError;
}

void 
CIRCode::EncodeValue(DWORD value, WORD &index, WORD length)
{
	for (int i = 0; i < length; i++)
	{
		if (value & 1)
			m_data[index].value = m_bpOne.value;
		else
			m_data[index].value = m_bpZero.value;
		index++;
		value >>= 1;
	}

}

BOOL CIRCode::IsKenwood(WORD length, WORD start, BOOL bIsRepeat)
{
	WORD	i;

	/* Kenwood format always has one leadin and one leadout pair. */
	if ((!bIsRepeat && length != 0x22) || 
		(bIsRepeat && m_CurrentFormat == IR_KENWOOD && length != 2))
		return FALSE;

	if (!bIsRepeat)
	{
		/* The interior is always all ones and zeros. */
		for (i = start + 1; i < start + 0x20; i++)
		{
			if (m_data[i].value != m_bpZero.value && m_data[i].value != m_bpOne.value)
				return FALSE;
		}


		/*
		 * See if the command code is properly complemented.
		 */
		for (i = start + 17; i < start + 25; i++)
		{
			if (m_data[i].value == m_bpZero.value && m_data[i + 8].value == m_bpZero.value ||
				m_data[i].value == m_bpOne.value && m_data[i + 8].value == m_bpOne.value)
				return FALSE;
		}
	}
	m_CurrentFormat = IR_KENWOOD;

	return TRUE;
}

void CIRCode::DecipherPairsKenwood(CCodePart &cp, WORD length, WORD start, BOOL bIsRepeat)
{
	if (bIsRepeat)
	{
		cp.m_nCommand = 0;
		cp.m_nDevice = 0;
		cp.m_bpPad.value = 0;
		cp.m_bpLeadIn.value = m_data[start].value;
		cp.m_bpLeadOut.value = m_data[start + 1].value;
	}
	else
	{

		cp.m_bpLeadIn.value = m_data[start].value;
		cp.m_nDevice = GrabInteger(start + 1, 16);
		cp.m_nCommand = GrabInteger(start + 17, 8);
		cp.m_bpLeadOut.value = m_data[start + 33].value;
		cp.m_bpPad.value = 0;
		cp.m_sRaw = _T("");
	}
}

int CIRCode::InterpretCodeKenwood(CCode &code)
{

	WORD	index = 0;
	DWORD	command = 0;
	DWORD	device = 0;

	/*
	 * Kenwood codes are always 34 pairs long.
	 */
	m_data.SetSize(68);

	/*
	 * Walk through the entries on the line and interpret each one.
	 */

	
	if (code.m_bOneTime)
	{
		m_data[index++].value = code.m_cpOneTime.m_bpLeadIn.value;
		device = (code.m_cpOneTime.m_nDevice) & 0xFFFF;
		command = (((~code.m_cpOneTime.m_nCommand) & 0xFF) << 8) | (code.m_cpOneTime.m_nCommand & 0xFF);
		EncodeValue(device, index, 16);
		EncodeValue(command, index, 16);
		m_data[index++].value = code.m_cpOneTime.m_bpLeadOut.value;
		code.m_cpOneTime.m_nLength = index;
	}

	if (code.m_bRepeat)
	{
		m_data[index++].value = code.m_cpRepeat.m_bpLeadIn.value;
		m_data[index++].value = code.m_cpRepeat.m_bpLeadOut.value;
		code.m_cpRepeat.m_nLength = index - code.m_cpOneTime.m_nLength;
	}
	return index;
}
