// IRCode.h: interface for the CIRCode class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_IRCODE_H__81E804E0_1220_11D3_AF56_00104B7981C1__INCLUDED_)
#define AFX_IRCODE_H__81E804E0_1220_11D3_AF56_00104B7981C1__INCLUDED_

//#include "..\struct.h"	// Added by ClassView
#include <afxtempl.h>
#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

const DWORD FUZZ_FACTOR = 4;

enum IRCODE_FORMAT
{
	IR_UNKNOWN,
	IR_DENON,
	IR_JVC,
	IR_KENWOOD,
	IR_NEC,
	IR_SONY_5,
	IR_SONY_8,
	IR_SONY_13
} ;


/*
 * During decoding, we maintain a histogram, sorted in descending order of
 * frequency, of burst pair values.
 */
typedef struct bp_histogram Histogram;
struct bp_histogram {
	DWORD	value;
	WORD	count;

	Histogram	*prev;
	Histogram	*next;
};


typedef union 
{
	WORD	pair[2];
	DWORD	value;
} CBurstPair;

typedef CArray<CBurstPair,CBurstPair&> pairArray;

typedef struct
{
	DWORD m_nDevice;
	DWORD m_nCommand;
	WORD	m_nLength;
	CBurstPair m_bpLeadIn;
	CBurstPair m_bpLeadOut;
	CBurstPair m_bpPad;
	CString m_sRaw;
} CCodePart;

typedef struct
{
	IRCODE_FORMAT m_type;
	WORD m_nCarrier;
	CBurstPair m_bpZero;
	CBurstPair m_bpOne;
	BOOL m_bOneTime;
	CCodePart m_cpOneTime;
	BOOL m_bRepeat;
	CCodePart m_cpRepeat;
} CCode;

class CIRCode  
{
public:
	int EncodeHex(CString& hex,  CCode& code);
	virtual int DecodeHex(const WORD hex[], WORD length, CCode& code);
	virtual int DecodeHex(LPCTSTR hex, CCode& code);
	virtual IRCODE_FORMAT WhatFormat();
	CIRCode();
	virtual ~CIRCode();

private:
	int InterpretCodeKenwood(CCode& code);
	void DecipherPairsKenwood(CCodePart& cp, WORD length, WORD start, BOOL bIsRepeat);
	BOOL IsKenwood(WORD length, WORD start, BOOL bIsRepeat);
	void EncodeValue(DWORD value, WORD & index, WORD length);
	void DecipherPairsRaw(CCodePart& cp, WORD length, WORD start, BOOL bIsRepeat);
	void DecipherPairsJVC(CCodePart& cp, WORD length, WORD start, BOOL bIsRepeat);
	void DecipherPairsSONY(CCodePart& cp, WORD length, WORD start, BOOL bIsRepeat);
	void DecipherPairsDENON(CCodePart& cp, WORD length, WORD start, BOOL bIsRepeat);
	void DecipherPairsNEC(CCodePart& cp, WORD length, WORD start, BOOL bIsRepeat);
	BOOL IsSONY(WORD length, WORD start, BOOL bIsRepeat);
	BOOL IsJVC(WORD length, WORD start, BOOL bIsRepeat);
	BOOL IsDENON(WORD length, WORD start, BOOL bIsRepeat);
	BOOL IsNEC(WORD length, WORD start, BOOL bIsRepeat);
	int DecipherPairs(CCodePart& cp, WORD length, WORD start, BOOL bIsRepeat);
	int InterpretCodeSONY(CCode& code);
	int InterpretCodeNEC(CCode& code);
	int InterpretCodeJVC(CCode& code);
	int InterpretCodeHEX(CCode& code, WORD start, BOOL bIsRepeat);
	int InterpretCodeDENON(CCode& code);
	int InterpretCode(CCode& code);
	void InterpretBit(LPCTSTR str, CBurstPair& val);
	int m_nError;
	DWORD GrabIntegerSony(int start, int length);
	DWORD GrabInteger(int start, int length);
	void FreeHistogram(Histogram* head);
	Histogram * CountValues();
	void SmoothValues();
	int CalculateBits();
	pairArray m_data;
	IRCODE_FORMAT m_CurrentFormat;
	CBurstPair m_bpZero;
	CBurstPair m_bpOne;
	WORD m_bySonyDeviceLength;
	WORD m_nRepeatLength;
	WORD m_nOneTimeLength;
	WORD m_nCarrier;
	CBurstPair m_bpRepeat;
	CBurstPair m_bpOneTime;
};

#endif // !defined(AFX_IRCODE_H__81E804E0_1220_11D3_AF56_00104B7981C1__INCLUDED_)
